@@ -1,4 +1,4 @@
-// $Id: xttpd.cc 5749 2014-10-11 19:42:10Z flaterco $
+// $Id: xttpd.cc 6225 2016-01-23 13:26:42Z flaterco $
 
 /*  xttpd  XTide web server.
 
@@ -35,6 +35,10 @@ using namespace libxtide;
 #include <pwd.h>
 #include <grp.h>
 
+#if systemd
+#include <systemd/sd-daemon.h>
+#endif
+
 
 // These browsers nowadays can get pretty verbose.
 static const size_t bufsize (10000);
@@ -64,56 +68,174 @@ static void checkedWrite (int fd, const void *buf, size_t count) {
 }
 
 
-static const bool parseAddress (unsigned short &portNumber,
-				unsigned long &addr,
-				char *arg) {
-  assert (arg);
-  // The : character is reserved for future use to distinguish
-  // IPv4 from IPv6 addresses.
-  if (strchr (arg, '/') || strchr (arg, '.')) {
-    char *c (strrchr (arg, '/'));
-    if (c) {
-      *c = '\0';
-      if (sscanf (c+1, "%hu", &portNumber) != 1)
-        portNumber = 80;
+static void setWebmaster () {
+  webmaster = getenv ("XTTPD_FEEDBACK");
+  #ifdef webmasteraddr
+    if (webmaster.isNull())
+      webmaster = webmasteraddr;
+  #endif
+  if (webmaster == "dave@flaterco.com")
+    webmaster = (char *)NULL;  // I can't handle the AOLers all by myself.
+  if (webmaster.isNull()) {
+    // Guess the local webmaster.  (Guessing the FQDN is harder!)
+    hostent *hp;
+    // On Suns, need struct keyword to avoid confusion with a global variable
+    // (extern struct utsname utsname declared in sys/utsname.h).
+    struct utsname foo;
+    uname (&foo);
+    if (!(hp = (hostent *) gethostbyname (foo.nodename)))
+      webmaster = foo.nodename;
+    else
+      webmaster = hp->h_name;
+    if (webmaster.strchr('.') == -1) {
+      // getdomainname() is useless
+      Dstr domain;
+      FILE *fp;
+      if ((fp = fopen ("/etc/defaultdomain", "r"))) {
+	domain.scan (fp);
+	fclose (fp);
+	webmaster += '.';
+	webmaster += domain;
+      } else if ((fp = fopen ("/etc/domain", "r"))) {
+	domain.scan (fp);
+	fclose (fp);
+	webmaster += '.';
+	webmaster += domain;
+      } else if ((fp = fopen ("/etc/resolv.conf", "r"))) {
+	Dstr buf, srch;
+	bool found=false;
+	while (!(buf.getline(fp)).isNull()) {
+	  buf /= domain;
+	  if (domain == "domain") {
+	    buf /= domain;
+	    webmaster += '.';
+	    webmaster += domain;
+	    found = true;
+	    break;
+	  } else if (srch.isNull() && domain == "search") {
+            buf /= srch;  // Just the first one.
+          }
+	}
+	fclose (fp);
+	if (!found && !srch.isNull()) {
+	  webmaster += '.';
+	  webmaster += srch;
+	}
+      }
     }
-    if (!inet_pton (AF_INET, arg, &addr)) {
-      Dstr details ("The offending address was ");
-      details += arg;
-      Global::barf (Error::BAD_IP_ADDRESS, details);
-    }
-    return true;
-  } else {
-    addr = INADDR_ANY;
-    if (sscanf (arg, "%hu", &portNumber) != 1) {
-      portNumber = 80;
-      return false;
-    }
-    return true;
+    if (webmaster %= "www.")
+      webmaster /= 4;
+    webmaster *= "webmaster@";
   }
+  webmaster *= "<a href=\"mailto:";
+  webmaster += "\">";
 }
 
 
-static void setupSocket (unsigned short portNumber,
-                         unsigned long addr,
-                         int &s) {
-  sockaddr_in sin;
-  hostent *hp;
-  // On Suns, need struct keyword to avoid confusion with a global variable
-  // (extern struct utsname utsname declared in sys/utsname.h).
-  struct utsname foo;
+#if !systemd
+static const bool parseAddress (sockaddr_storage &addr,
+				constString arg) {
+  assert (arg);
 
-  uname (&foo);
-  if (!(hp = (hostent *) gethostbyname (foo.nodename))) {
-    Dstr details ("Gethostbyname fails on own nodename!");
-    Global::barf (Error::CANT_GET_SOCKET, details);
+  // getaddrinfo parses addresses but not port numbers.
+  // 1. If it starts with hyphen then it's a setting.
+  // 2. If it's a null string or a single slash, reject.
+  // 3. If it contains a slash then parse as address/port.
+  // 4. If it's entirely numeric then parse as port.
+  // 5. Else parse as address or hostname.
+
+  bool gotAddr=false, gotPort=false;
+  unsigned short portNumber;
+  Dstr darg(arg);
+  char *caddr = NULL;
+
+  if (darg[0] == '-')
+    return false;
+  else if (darg.length()==0 || darg == "/")
+    Global::barf (Error::PORT_FAIL);
+  else if (darg.contains('/')) {
+    // Address/port
+    int i = darg.strrchr('/');
+    assert (i >= 0);
+    if ((unsigned)i+1 < darg.length()) {     // Maybe they just said addr/
+      if (sscanf (darg.ascharfrom(i+1), "%hu", &portNumber) == 1)
+        gotPort = true;
+      else
+        Global::barf (Error::PORT_FAIL);
+    }
+    if (i > 0) {                             // Maybe they just said /port
+      darg -= i;
+      caddr = darg.aschar();
+      gotAddr = true;
+    }
+  } else {
+    bool isNum = true;
+    for (unsigned i=0; i<darg.length(); ++i) {
+      const char c = darg[i];
+      if (c < '0' || c > '9') {
+        isNum = false;
+        break;
+      }
+    }
+    if (isNum) {
+      // Port number
+      if (sscanf (arg, "%hu", &portNumber) == 1)
+        gotPort = true;
+      else
+        Global::barf (Error::PORT_FAIL);
+    } else {
+      // Address or host name
+      caddr = darg.aschar();
+      gotAddr = true;
+    }
   }
-  memset ((char *) &sin, '\0', sizeof sin);
-  memcpy ((char *) &sin.sin_addr, hp->h_addr, hp->h_length);
-  sin.sin_port = htons (portNumber);
-  sin.sin_family = hp->h_addrtype;
-  sin.sin_addr.s_addr = addr;
-  if ((s = socket (hp->h_addrtype, SOCK_STREAM, 0)) == -1) {
+
+  if (!gotPort)
+    portNumber = 80;
+
+  if (gotAddr) {
+    struct addrinfo hint, *res;
+    memset(&hint, 0, sizeof hint);
+    hint.ai_family   = PF_UNSPEC;
+    hint.ai_socktype = SOCK_STREAM;
+    hint.ai_protocol = IPPROTO_TCP;
+    int ret = getaddrinfo(caddr, NULL, &hint, &res);
+    if (ret) {
+      Dstr details ("getaddrinfo: ");
+      details += gai_strerror(ret);
+      details += "\nThe offending address was ";
+      details += caddr;
+      Global::barf (Error::ADDRESS_FAIL, details);
+    }
+    assert (res);
+    assert (res->ai_addrlen <= sizeof(addr));
+    memcpy (&addr, res->ai_addr, res->ai_addrlen);
+    freeaddrinfo (res);
+    switch (addr.ss_family) {
+    case AF_INET:
+      ((sockaddr_in *)&addr)->sin_port = htons(portNumber);
+      break;
+    case AF_INET6:
+      ((sockaddr_in6 *)&addr)->sin6_port = htons(portNumber);
+      break;
+    default:
+      Global::barf (Error::CANT_GET_SOCKET, "Surprising family in parseAddress.");
+    }
+  } else {
+    memset (&addr, 0, sizeof(addr));
+    sockaddr_in *sin = (sockaddr_in *)&addr;
+    sin->sin_family      = AF_INET;
+    sin->sin_port        = htons(portNumber);
+    sin->sin_addr.s_addr = INADDR_ANY;
+  }
+
+  return true;
+}
+
+
+static void setupSocket (sockaddr_storage &addr,
+                         int &s) {
+  if ((s = socket (addr.ss_family, SOCK_STREAM, 0)) == -1) {
     Dstr details ("socket: ");
     details += strerror (errno);
     details += '.';
@@ -136,13 +258,48 @@ static void setupSocket (unsigned short portNumber,
       Global::xperror ("setsockopt");
     }
   }
-  if (bind (s, (sockaddr *)(&sin), sizeof sin) == -1) {
-    if (errno == EADDRINUSE) {
+  bool bound = (bind(s,(sockaddr *)&addr,sizeof(addr))==0);
+  int yerrno = errno;
+  if (!bound && addr.ss_family == AF_INET6) {
+
+    // Here comes trouble.
+
+    // The Linux kernel (bind syscall) will sometimes refuse to bind an
+    // address (likely a link-local address) unless it is told which network
+    // interface to bind it on.  Specifically, it expects you to put the
+    // interface index in the sin6_scope_id field of the sockaddr_in6.  Apps
+    // are now supposed to implement the convoluted process of looking up the
+    // interface index given an address (see example at
+    // http://stackoverflow.com/questions/13504934/binding-sockets-to-ipv6-addresses)
+    // or else require the user to specify a network interface.
+
+    // Network interface indices are just serial numbers that are assigned
+    // starting at 1.  Typically, index 1 will be loopback, index 2 will be
+    // the first or only real network interface, and any extra interfaces
+    // will be index 3 and up.
+
+    // The scope ID is a *different* magic number (see
+    // https://en.wikipedia.org/wiki/IPv6_address#IPv6_address_scopes).  It
+    // so happens that the scope ID corresponding to link-local scope is 2.
+    // 3 is not a defined scope ID, but 4 and 5 are.  So we have here a
+    // recipe for truly epic confusion:  putting the scope ID in the scope ID
+    // field will work a lot of the time but not always, and the error codes
+    // from bind are nonspecific about the problem.
+
+    for (unsigned magic=4; !bound && magic>0; --magic) {
+      ((sockaddr_in6 *)&addr)->sin6_scope_id = magic;
+      bound = (bind(s,(sockaddr *)&addr,sizeof(addr))==0);
+    }
+  }
+  // Report the error from the first failure so that it won't be obscured by
+  // other errors caused by the magic number guessing.
+  if (!bound) {
+    if (yerrno == EADDRINUSE) {
       Dstr details ("Socket already in use.");
       Global::barf (Error::CANT_GET_SOCKET, details);
     } else {
       Dstr details ("bind: ");
-      details += strerror (errno);
+      details += strerror (yerrno);
       details += '.';
       Global::barf (Error::CANT_GET_SOCKET, details);
     }
@@ -154,52 +311,8 @@ static void setupSocket (unsigned short portNumber,
     details += '.';
     Global::barf (Error::CANT_GET_SOCKET, details);
   }
-
-  webmaster = getenv ("XTTPD_FEEDBACK");
-#ifdef webmasteraddr
-  if (webmaster.isNull())
-    webmaster = webmasteraddr;
-#endif
-  if (webmaster == "dave@flaterco.com")
-    webmaster = (char *)NULL;  // I can't handle the AOLers all by myself.
-  if (webmaster.isNull()) {
-    // Guess the local webmaster.  (Guessing the FQDN is harder!)
-    webmaster = hp->h_name;
-    if (webmaster.strchr('.') == -1) {
-      // They deleted getdomainname()...  Now what am I supposed to do?
-      Dstr domain;
-      FILE *fp;
-      if ((fp = fopen ("/etc/defaultdomain", "r"))) {
-	domain.scan (fp);
-	fclose (fp);
-	webmaster += '.';
-	webmaster += domain;
-      } else if ((fp = fopen ("/etc/domain", "r"))) {
-	domain.scan (fp);
-	fclose (fp);
-	webmaster += '.';
-	webmaster += domain;
-      } else if ((fp = fopen ("/etc/resolv.conf", "r"))) {
-	Dstr buf;
-	while (!(buf.getline(fp)).isNull()) {
-	  buf /= domain;
-	  if (domain == "domain") {
-	    buf /= domain;
-	    webmaster += '.';
-	    webmaster += domain;
-	    break;
-	  }
-	}
-	fclose (fp);
-      }
-    }
-    if (webmaster %= "www.")
-      webmaster /= 4;
-    webmaster *= "webmaster@";
-  }
-  webmaster *= "<a href=\"mailto:";
-  webmaster += "\">";
 }
+#endif
 
 
 static void addDisclaimer (Dstr &foo) {
@@ -1020,27 +1133,26 @@ you should check the indexes linked below.</p>\n";
 }
 
 
-static void logHits (const sockaddr *addr) {
-  Dstr msg;
-  switch (addr->sa_family) {
+// 2016-01-10  Rest in peace, gethostbyaddr.
+static void logHits (const sockaddr *saddr, acceptarg3_t saddrlen) {
+  char name[256], caddr[256];
+  void *src = NULL;
+  strcpy(caddr, "?");
+  switch (saddr->sa_family) {
   case AF_INET:
-    {
-      sockaddr_in *u ((sockaddr_in *)addr);
-      unsigned long a (u->sin_addr.s_addr);
-      // Archaic SunOS syntax: inet_ntoa (&a);
-      char *c (inet_ntoa (u->sin_addr));
-      msg = c;
-      msg += "  ";
-      hostent *hp;
-      if ((hp = gethostbyaddr ((char *)(&a), sizeof a, AF_INET)) == NULL)
-        msg += "(?)";
-      else
-        msg += hp->h_name;
-    }
+    src = &(((sockaddr_in *)saddr)->sin_addr);
+    break;
+  case AF_INET6:
+    src = &(((sockaddr_in6 *)saddr)->sin6_addr);
     break;
-  default:
-    msg = "Non-Internet address (?)";
   }
+  if (src)
+    (void)inet_ntop(saddr->sa_family, src, caddr, sizeof(caddr));
+  if (getnameinfo(saddr, saddrlen, name, sizeof(name), NULL, 0, NI_NAMEREQD))
+    strcpy(name, "(?)");
+  Dstr msg (caddr);
+  msg += "  ";
+  msg += name;
   Global::log (msg, LOG_INFO);
 }
 
@@ -1060,7 +1172,9 @@ Disallow: /\n");
 }
 
 
-static void handleRequest (const sockaddr *addr, int s) {
+static void handleRequest (const sockaddr *addr,
+			   acceptarg3_t addrlen,
+			   int s) {
   // It would have been kind of lame to do this in main and make every
   // request get the same random sequence.
   srand (time (NULL));
@@ -1068,7 +1182,7 @@ static void handleRequest (const sockaddr *addr, int s) {
   char request[bufsize+1];
   Dstr buf, command, filename;
 
-  logHits (addr);
+  logHits (addr, addrlen);
 
   // This will truncate long requests, which is better than they deserve.
   ssize_t len (read (s, request, bufsize));
@@ -1080,6 +1194,10 @@ static void handleRequest (const sockaddr *addr, int s) {
   }
   request[len] = '\0';
   buf = request;
+  if (buf.length() == 0) {
+    Global::log ("Empty client request", LOG_NOTICE);
+    badRequestBarf (s, "Empty request");
+  }
 
   buf /= command;
   if (command != "GET") {
@@ -1155,17 +1273,6 @@ static void dontBeRoot() {
 
 
 int main (int argc, char **argv) {
-
-  // To run on console, deactivate this block.
-  Global::setDaemonMode();
-  pid_t fpid (fork());
-  if (fpid == -1) {
-    Global::xperror ("fork");
-    exit (-1);
-  }
-  if (fpid)
-    exit (0);
-
   int listenSocket;
 
   // The port number given to xttpd is not a "proper" command line
@@ -1173,15 +1280,48 @@ int main (int argc, char **argv) {
   // if we get one we hide it.
   bool portNumberEvade (false);
 
-  // Bind the port and then drop root immediately.
-  {
-    unsigned short portnum (80);
-    unsigned long addr (INADDR_ANY);
-    if (argc >= 2)
-      portNumberEvade = parseAddress (portnum, addr, argv[1]);
-    setupSocket (portnum, addr, listenSocket);
-  }
+  #if !XTTPD_NO_DAEMON
+    Global::setDaemonMode();
+  #endif
+
+  #if systemd
+    {
+      int sderr = sd_listen_fds(0);
+      if (sderr < 0) {
+	Dstr details ("sd_listen_fds: ");
+	details += strerror (errno);
+	details += '.';
+	Global::barf (Error::SYSTEMD_NO_SOCKET, details);
+      } else if (sderr != 1)
+	Global::barf (Error::SYSTEMD_NO_SOCKET, "Wrong number of file descriptors received from systemd.");
+    }
+    listenSocket = SD_LISTEN_FDS_START;
+  #else
+    #if !XTTPD_NO_DAEMON
+      if (daemon(0,0) == -1) {
+        Global::xperror ("daemon");
+        exit (-1);
+      }
+    #endif
+
+    // Bind the port and then drop root immediately.
+    {
+      sockaddr_storage addr;
+      if (argc >= 2)
+	portNumberEvade = parseAddress (addr, argv[1]);
+      if (!portNumberEvade) {
+	// Default address and port
+	memset (&addr, 0, sizeof(addr));
+	sockaddr_in *sin = (sockaddr_in *)&addr;
+	sin->sin_family      = AF_INET;
+	sin->sin_port        = htons(80);
+	sin->sin_addr.s_addr = INADDR_ANY;
+      }
+      setupSocket (addr, listenSocket);
+    }
+  #endif
   dontBeRoot();
+  setWebmaster();
 
   // An unfortunate consequence of needing to drop root ASAP is that
   // failure to bind the port prevents xttpd -v from working.
@@ -1196,7 +1336,7 @@ int main (int argc, char **argv) {
 
   // This should never be used (Visual Studio 2012 can't build xttpd).
   #ifdef NO_INITIALIZER_LISTS
-  Global::initializeGraphFont();
+    Global::initializeGraphFont();
   #endif
 
   // Build the indices.
@@ -1207,11 +1347,15 @@ int main (int argc, char **argv) {
     zoneinfoIsNotHorriblyObsolete = nowarn.zoneinfoIsNotHorriblyObsolete();
   }
   Global::log ("Accepting connections.", LOG_NOTICE);
+  #if systemd
+    sd_notify(0, "READY=1");
+  #endif
 
   fd_set rdset;
   FD_ZERO (&rdset);
   FD_SET (listenSocket, &rdset);
-  sockaddr addr;
+  sockaddr_storage addr;
+  sockaddr *addrp = (sockaddr *)&addr;
   while (true) {
     fd_set trdset, twrset, texset;
     trdset = rdset;
@@ -1222,12 +1366,14 @@ int main (int argc, char **argv) {
     int ts (-1);
     ts = select (listenSocket+1, &trdset, &twrset, &texset, NULL);
     if (ts > 0) {
-      // Type of length is set by configure.
-      acceptarg3_t length (sizeof (addr));
+      // acceptarg3_t is set by configure, but now that we are relying on
+      // new-age net functions, it pretty much has to be socklen_t.
+      acceptarg3_t length (sizeof(addr));
       int newSocket;
-      if ((newSocket = accept (listenSocket, &addr, &length)) != -1) {
+      if ((newSocket = accept (listenSocket, addrp, &length)) != -1) {
+	assert (length <= sizeof(addr));
         if (!fork())
-          handleRequest (&addr, newSocket);
+          handleRequest (addrp, length, newSocket);
         close (newSocket);
       }
     }
@@ -1247,5 +1393,3 @@ int main (int argc, char **argv) {
 // improve this code, and in the event that serious problems are found
 // with it (like security vulnerabilities), the most likely treatment
 // will be euthanasia.
-
-// Cleanup2006 Legacy Cruft CloseEnough
